/*
 *		linux/arch/m68k/amiga/zorro.c
 *
 *		Copyright (C) 1995 Geert Uytterhoeven
 *
 *		This file is subject to the terms and conditions of the GNU General Public
 *		License.  See the file README.legal in the main directory of this archive
 *		for more details.
 */


#include <linux/types.h>
#include <linux/kernel.h>

#include <asm/bootinfo.h>
#include <asm/amigatypes.h>


	/*
	 *		Zorro Expansion Device Manufacturers and Products
	 */

struct Manufacturer {
	char *Name;
	u_short ID;
	u_short NumProd;
	struct Product *Products;
};

struct Product {
	char *Name;
	u_char ID;
};

#define DEF_PROD(name, id) \
	{ name, PROD_##id }
#define DEF_MANUF(name, id) \
	{ name, MANUF_##id, sizeof(Prod_##id)/sizeof(struct Product), Prod_##id }


	/*
	 *		Known Zorro Expansion Devices
	 *
	 *		Warning: Make sure the Manufacturer and Product names are not too
	 *					long (max. 80 characters per board indentification line)
	 */

static struct Product Prod_COMMODORE[] = {
	DEF_PROD("A2090/A2090A HD Controller", A2090A),
	DEF_PROD("A2091 HD Controller", A2091),
	DEF_PROD("A590 HD Controller", A590),
	DEF_PROD("A2090B 2090 Autoboot Card", A2090B),
	DEF_PROD("Arcnet Card", ARCNET),
	DEF_PROD("A2052/58.RAM | 590/2091.RAM", CBMRAM),
	DEF_PROD("A560 Memory Module", A560RAM),
	DEF_PROD("A2232 Serial Prototype", A2232PROTO),
	DEF_PROD("A2232 Serial Production", A2232),
	DEF_PROD("A2620 68020/RAM Card", A2620),
	DEF_PROD("A2630 68030/RAM Card", A2630),
	DEF_PROD("A4091 SCSI HD Controller", A4091),
	DEF_PROD("Romulator Card", ROMULATOR),
	DEF_PROD("A3000 Test Fixture", A3000TESTFIX),
	DEF_PROD("A2065 Ethernet Card", A2065)
};

static struct Product Prod_COMMODORE2[] = {
	DEF_PROD("Bridgeboard", BRIDGEBOARD),
	DEF_PROD("A2386-SX Bridgeboard", A2386SX),
};

static struct Product Prod_U_LOWELL[] = {
	DEF_PROD("A2410 Hires Graphics Card", A2410)
};

static struct Product Prod_AMERISTAR[] = {
	DEF_PROD("A2065 Ethernet Card", AMERISTAR2065)
};

static struct Product Prod_GVP[] = {
	DEF_PROD("Series II SCSI Controller", GVPIISCSI),
	DEF_PROD("Series II SCSI Controller", GVPIISCSI_2),
	DEF_PROD("GVP Series II RAM", GVPIIRAM),
	DEF_PROD("GVP IV-24 Graphics Board", GVPIV_24),
	DEF_PROD("GVP I/O Extender", GVPIO_EXT)
};

static struct Product Prod_GVP2[] = {
	DEF_PROD("GVP Spectrum Graphics Board (RAM)", SPECTRUM_RAM),
	DEF_PROD("GVP Spectrum Graphics Board (REG)", SPECTRUM_REG)
};

static struct Product Prod_VILLAGE_TRONIC[] = {
	DEF_PROD("Ariadne Ethernet Card", ARIADNE),
	DEF_PROD("Picasso II Graphics Board (RAM)", PICASSO_II_RAM),
	DEF_PROD("Picasso II Graphics Board (REG)", PICASSO_II_REG)
};

static struct Product Prod_HELFRICH1[] = {
	DEF_PROD("Rainbow3 Graphics Board", RAINBOW3)
};

static struct Product Prod_HELFRICH2[] = {
	DEF_PROD("Piccolo Graphics Board (RAM)", PICCOLO_RAM),
	DEF_PROD("Piccolo Graphics Board (REG)", PICCOLO_REG),
	DEF_PROD("PeggyPlus MPEG Decoder Board", PEGGY_PLUS),
	DEF_PROD("SD64 Graphics Board (RAM)", SD64_RAM),
	DEF_PROD("SD64 Graphics Board (REG)", SD64_REG)
};

static struct Product Prod_HYDRA_SYSTEMS[] = {
	DEF_PROD("Amiganet Board", AMIGANET)
};

static struct Product Prod_MEMPHIS[] = {
	DEF_PROD("Stormbringer", STORMBRINGER)
};

static struct Product Prod_SUPRA[] = {
	DEF_PROD("Supra 2400 Modem", SUPRA_2400MODEM)
};

static struct Product Prod_BSC[] = {
	DEF_PROD("BSC ALF 3 SCSI Controller", ALF_3_SCSI)
};

static struct Product Prod_BSC2[] = {
	DEF_PROD("BSC Oktagon 2008 SCSI Controller", OKTAGON_2008),
	DEF_PROD("BSC Tandem", TANDEM),
	DEF_PROD("BSC Par/Ser Card", PAR_SER_CARD),
	DEF_PROD("BSC MultiFace III", MULTIFACE_III),
	DEF_PROD("BSC ISDN Master", ISDN_MASTER)
};

static struct Product Prod_JOCHHEIM[] = {
	DEF_PROD("Jochheim RAM", JOCHHEIM_RAM)
};

static struct Product Prod_CHECKPOINT[] = {
	DEF_PROD("Serial Solution", SERIAL_SOLUTION)
};

static struct Product Prod_GOLEM[] = {
	DEF_PROD("Golem SCSI-II Controller", GOLEM_SCSI_II)
};

static struct Product Prod_ADV_SYS_SOFT[] = {
	DEF_PROD("Nexus SCSI Controller", NEXUS_SCSI),
	DEF_PROD("Nexus RAM", NEXUS_RAM)
};

static struct Product Prod_XPERT_PRODEV[] = {
	DEF_PROD("Merlin Graphics Board (RAM)", MERLIN_RAM),
	DEF_PROD("Merlin Graphics Board (REG)", MERLIN_REG)
};

static struct Product Prod_UTILITIES_ULTD[] = {
	DEF_PROD("Emplant Deluxe", EMPLANT_DELUXE),
	DEF_PROD("Emplant Deluxe", EMPLANT_DELUXE2)
};

static struct Product Prod_HARMS_PROF[] = {
	DEF_PROD("3500 Turbo board", 3500_TURBO)
};

static struct Product Prod_VORTEX[] = {
	DEF_PROD("Golden Gate 80386 Board", GOLDEN_GATE_386),
	DEF_PROD("Golden Gate RAM", GOLDEN_GATE_RAM),
	DEF_PROD("Golden Gate 80486 Board", GOLDEN_GATE_486)
};

static struct Product Prod_PHASE5[] = {
	DEF_PROD("FastLane RAM", FASTLANE_RAM),
	DEF_PROD("FastLane SCSI", FASTLANE_SCSI),
	DEF_PROD("CyberVision64 Graphics Board", CYBERVISION)
};

static struct Product Prod_APOLLO[] = {
	DEF_PROD("AT-Apollo", AT_APOLLO)
};

static struct Product Prod_UWE_GERLACH[] = {
	DEF_PROD("RAM/ROM", UG_RAM_ROM)
};

static struct Product Prod_MACROSYSTEMS[] = {
	DEF_PROD("Warp Engine", WARP_ENGINE)
};

static struct Product Prod_MACROSYSTEMS2[] = {
	DEF_PROD("Maestro", MAESTRO),
	DEF_PROD("VLab", VLAB),
	DEF_PROD("Maestro Pro", MAESTRO_PRO),
	DEF_PROD("Retina Z2 Graphics Board", RETINA_Z2),
	DEF_PROD("MultiEvolution", MULTI_EVOLUTION)
};

static struct Product Prod_MICROBOTICS[] = {
	DEF_PROD("VXL-30 Turbo Board", VXL_30)
};

static struct Manufacturer Manufacturers[] = {
	DEF_MANUF("Commodore", COMMODORE),
	DEF_MANUF("Commodore", COMMODORE2),
	DEF_MANUF("University of Lowell", U_LOWELL),
	DEF_MANUF("Ameristar", AMERISTAR),
	DEF_MANUF("Helfrich", HELFRICH1),
	DEF_MANUF("Helfrich", HELFRICH2),
	DEF_MANUF("Great Valley Products", GVP),
	DEF_MANUF("Great Valley Products", GVP2),
	DEF_MANUF("Village Tronic", VILLAGE_TRONIC),
	DEF_MANUF("Hydra Systems", HYDRA_SYSTEMS),
	DEF_MANUF("Memphis", MEMPHIS),
	DEF_MANUF("Supra", SUPRA),
	DEF_MANUF("BSC", BSC),
	DEF_MANUF("BSC", BSC2),
	DEF_MANUF("Jochheim", JOCHHEIM),
	DEF_MANUF("Checkpoint Technologies", CHECKPOINT),
	DEF_MANUF("Golem", GOLEM),
	DEF_MANUF("Advanced Systems & Software", ADV_SYS_SOFT),
	DEF_MANUF("XPert/ProDev", XPERT_PRODEV),
	DEF_MANUF("Utilities Unlimited", UTILITIES_ULTD),
	DEF_MANUF("Harms Professional", HARMS_PROF),
	DEF_MANUF("Vortex", VORTEX),
	DEF_MANUF("Phase5", PHASE5),
	DEF_MANUF("Apollo", APOLLO),
	DEF_MANUF("Uwe Gerlach", UWE_GERLACH),
	DEF_MANUF("MacroSystems", MACROSYSTEMS),
	DEF_MANUF("MacroSystems", MACROSYSTEMS2),
	DEF_MANUF("MicroBotics", MICROBOTICS)
};

#define NUM_MANUF (sizeof(Manufacturers)/sizeof(struct Manufacturer))


	/*
	 *		Identify an AutoConfig Expansion Device
	 *
	 *		If the board was configured by a Linux/68k driver, an asterisk will
	 *		be printed before the board address (except for unknown and `Hacker
	 *		Test' boards).
	 */

static int identify(int devnum, char *buf)
{
	struct ConfigDev *cd;
	int manuf, prod;
	u_long addr, size;
	char *manufname, *prodname;
	char zorro, mag, configured;
	int identified = 0;
	int i, j, len = 0;

	cd = &boot_info.bi_amiga.autocon[devnum];
	manuf = cd->cd_Rom.er_Manufacturer;
	prod = cd->cd_Rom.er_Product;
	addr = (u_long)cd->cd_BoardAddr;
	size = cd->cd_BoardSize;
	configured = boot_info.bi_amiga.autocon_configured & (1<<devnum) ? '*' : ' ';
	manufname = prodname = "<UNKNOWN>";

	for (i = 0; i < NUM_MANUF; i++)
		if (Manufacturers[i].ID == manuf) {
			manufname = Manufacturers[i].Name;
			for (j = 0; j < Manufacturers[i].NumProd; j++)
				if (Manufacturers[i].Products[j].ID == prod) {
					prodname = Manufacturers[i].Products[j].Name;
					identified = 1;
					break;
				}
			break;
		}

	switch (cd->cd_Rom.er_Type & ERT_TYPEMASK) {
		case ERT_ZORROII:
			zorro = '2';
			break;
		case ERT_ZORROIII:
			zorro = '3';
			break;
		default:
			zorro = '?';
			break;
	}
	if (size & 0xfffff) {
		size >>= 10;
		mag = 'K';
	} else {
		size >>= 20;
		mag = 'M';
	}

	if (identified)
		len = sprintf(buf, " %c0x%08lx: %s %s (Z%c, %ld%c)\n", configured, addr,
						  manufname, prodname, zorro, size, mag);
	else if (manuf == MANUF_HACKER)
		len = sprintf(buf, "  0x%08lx: Hacker Test Board 0x%02x (Z%c, %ld%c)\n",
						  addr, prod, zorro, size, mag);
	else {
		len = sprintf(buf, "  0x%08lx: [%04x:%02x] made by %s (Z%c, %ld%c)\n",
						  addr, manuf, prod, manufname, zorro, size, mag);
		len += sprintf(buf+len, "  Please report this unknown device to "
							"Geert.Uytterhoeven@cs.kuleuven.ac.be\n");
	}
	return(len);
}


	/*
	 *		Identify all known AutoConfig Expansion Devices
	 */

void amiga_zorro_ident(void)
{
	int i;
	char tmp[160];

	printk("Probing AutoConfig expansion device(s):\n");
	for (i = 0; i < boot_info.bi_amiga.num_autocon; i++) {
		identify(i, tmp);
		printk(tmp);
	}
	if (!boot_info.bi_amiga.num_autocon)
		printk("No AutoConfig expansion devices present.\n");
}

int get_zorro_list(char *buffer)
{
	int i, j, len = 0;
	char tmp[160];

	if (boot_info.machtype == MACH_AMIGA) {
		len = sprintf(buffer, "AutoConfig expansion devices:\n");
		for (i = 0; i < boot_info.bi_amiga.num_autocon; i++) {
			j = identify(i, tmp);
			if (len+j >= 4075) {
				len += sprintf(buffer+len, "4K limit reached!\n");
				break;
			}
			strcpy(buffer+len, tmp);
			len += j;
		}
	}
	return(len);
}
