/*
 *  linux/arch/m68k/kernel/setup.c
 *
 *  Copyright (C) 1995  Hamish Macdonald
 *
 *  95/11/18 Richard Hirst: Added support for MVME166
 */

/*
 * This file handles the architecture-dependent parts of system setup
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/console.h>
#include <linux/genhd.h>
#include <linux/errno.h>

#include <asm/bootinfo.h>

struct bootinfo boot_info = {0,};
int bisize = sizeof boot_info;

int m68k_is040 = 0;

extern int ramdisk_size;
extern int end;
extern unsigned long availmem;

/* setup some dummy routines */
static unsigned long dummy_floppy_init (unsigned long start,
					unsigned long end)
{
    return start;
}
static int dummy_floppy_change (struct buffer_head *bh)
{
    return 0;
}
static unsigned long dummy_hd_init (unsigned long start,
				    unsigned long end)
{
    return start;
}
static void dummy_waitbut(void)
{
}
static void dummy_hd_setup (char *str, int *ints)
{
}

void (*mach_sched_init) (isrfunc);
unsigned long (*mach_keyb_init) (unsigned long);
int (*mach_kbdrate) (struct kbd_repeat *) = NULL;
void (*mach_init_INTS) (void);
int (*mach_add_isr) (unsigned long, isrfunc, int, void *);
void (*mach_process_int) (int, struct intframe *) = NULL;
unsigned long (*mach_gettimeoffset) (void);
void (*mach_gettod) (int*, int*, int*, int*, int*, int*);
int (*mach_hwclk) (int, struct hwclk_time*) = NULL;
int (*mach_set_clock_mmss) (unsigned long) = NULL;
void (*mach_check_partition) (struct gendisk *, unsigned int);
void (*mach_mksound)( unsigned int count, unsigned int ticks );
void (*mach_reset)( void );
unsigned long (*mach_floppy_init) (unsigned long, unsigned long) = dummy_floppy_init;
int (*mach_floppy_change) (struct buffer_head *bh) = dummy_floppy_change;
unsigned long (*mach_hd_init) (unsigned long, unsigned long) = dummy_hd_init;
void (*waitbut)(void) = dummy_waitbut;
void (*mach_hd_setup) (char *, int *) = dummy_hd_setup;

extern void config_amiga(void);
extern void config_atari(void);
extern void config_mac(void);
extern void config_mvme166(void);

extern void register_console(void (*proc)(const char *));
extern void ami_serial_print (const char *str);
extern void ata_serial_print (const char *str);
extern void mvme166_serial_print (const char *str);

#define MASK_256K 0xfffc0000

void setup_arch(char **cmdline_p,
	unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	unsigned long memory_start, memory_end;
	int i;

#ifdef CONFIG_AMIGA
	if (boot_info.machtype == MACH_AMIGA)
	  register_console (ami_serial_print);
#endif
#ifdef CONFIG_ATARI
	if (boot_info.machtype == MACH_ATARI)
	  register_console (ata_serial_print);
#endif
#ifdef xxCONFIG_MVME166
	if (boot_info.machtype == MACH_MVME166)
	  register_console (mvme166_serial_print);
#endif

	ramdisk_size = boot_info.ramdisk_size;
	if (boot_info.cputype & CPU_68040)
		m68k_is040 = 1;

	memory_start = availmem;
	memory_end = 0;

	for (i = 0; i < boot_info.num_memory; i++)
		memory_end += boot_info.memory[i].size & MASK_256K;

	*cmdline_p = boot_info.command_line;
	*memory_start_p = memory_start;
	*memory_end_p = memory_end;

	switch (boot_info.machtype) {
#ifdef CONFIG_AMIGA
	    case MACH_AMIGA:
		config_amiga();
		break;
#endif
#ifdef CONFIG_ATARI
	    case MACH_ATARI:
		config_atari();
		break;
#endif
#ifdef CONFIG_MAC
	    case MACH_MAC:
		config_mac();
		break;
#endif
#ifdef CONFIG_MVME166
	    case MACH_MVME166:
		config_mvme166();
		break;
#endif
	    default:
		panic ("No configuration setup");
	}
}

unsigned char getledstate(void)
{
	return 0;
}

struct kbd_struct;
void setledstate(struct kbd_struct *kbd, unsigned int led)
{
}

int setkeycode(unsigned int scancode, unsigned int keycode)
{
	return -EOPNOTSUPP;
}

int getkeycode(unsigned int scancode)
{
	return -EOPNOTSUPP;
}

void floppy_setup(char *str, int *ints)
{
}
