/*
 * arch/m68k/mvme166/166part.c
 *
 * Richard Hirst [richard@sleepie.demon.co.uk]
 *
 * Based on linux/amiga/amipart.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>

#define SWAP_INT(x)	(((unsigned char)x << 24) | \
			 ((unsigned char)(x >> 8) << 16) | \
			 ((unsigned char)(x >> 16) << 8) | \
			 ((unsigned char)(x >> 24)))

extern int current_minor;

static char minor_name (struct gendisk *hd, int minor)
{
	char base_name = (hd->major == IDE1_MAJOR) ? 'c' : 'a';
	return base_name + (minor >> hd->minor_shift);
}

static void add_partition (struct gendisk *hd, int minor, int start, int size)
{
	hd->part[minor].start_sect = start;
	hd->part[minor].nr_sects   = size;
	printk(" %s%c%d", hd->major_name, minor_name(hd, minor),
		minor & ((1 << hd->minor_shift) - 1));
}

void mvme166_check_partition(struct gendisk *hd, unsigned int dev)
{
	static int first_time = 1;
	int i, minor = current_minor;
	struct buffer_head *bh;
	struct partition *p;
	unsigned long first_sector;

	if (first_time)
		printk("Partition check:\n");
	first_time = 0;
	first_sector = hd->part[MINOR(dev)].start_sect;

	/*
	 * This is a kludge to allow the partition check to be
	 * skipped for specific drives (ie. IDE cd-rom drives)
	 */
	if ((int)first_sector == -1) {
		hd->part[MINOR(dev)].start_sect = 0;
		return;
	}

	if (!(bh = bread(dev,0,1024))) {
		printk("  unable to read partition table of device %04x\n",dev);
		return;
	}
	printk("  %s%c:", hd->major_name, minor_name(hd, minor));
	current_minor += 4;  /* first "extra" minor (for extended partitions) */
	if (*(unsigned short *) (bh->b_data+510) == 0x55AA) {
		p = (struct partition *) (0x1BE + bh->b_data);
		for (i=1 ; i<=4 ; minor++,i++,p++) {
			if (!p->nr_sects)
				continue;
			add_partition(hd, minor,
				first_sector + SWAP_INT(p->start_sect),
					SWAP_INT(p->nr_sects));
			if ((current_minor & 0x3f) >= 60)
				continue;
#if 0
			if (p->sys_ind == EXTENDED_PARTITION) {
				printk(" <");
				extended_partition(hd, (hd->major << 8) | minor);
				printk(" >");
			}
#endif
		}
	} else
		printk(" bad partition table");
	printk("\n");
	brelse(bh);
}

