/*
 *  arch/m68k/mvme166/config.c
 *
 *  Copyright (C) 1995 Richard Hirst [richard@sleepie.demon.co.uk]
 *
 * Based on:
 *
 *  linux/amiga/config.c
 *
 *  Copyright (C) 1993 Hamish Macdonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

#include <stdarg.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/kd.h>
#include <linux/tty.h>
#include <linux/console.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/bootinfo.h>
#include <asm/irq.h>

typedef struct {
	unsigned char
		ctrl,
		bcd_sec,
		bcd_min,
		bcd_hr,
		bcd_dow,
		bcd_dom,
		bcd_mth,
		bcd_year;
} MK48T08;

#define RTC_WRITE	0x80
#define RTC_READ	0x40
#define RTC_STOP	0x20

MK48T08 * volatile rtc = (MK48T08 *)0xfffc1ff8;

extern void mvme166_serial_print (const char *str);
extern void mvme166_process_int (int level, struct intframe *fp);
extern void mvme166_init_INTS (void);
extern void mvme166_sched_init(isrfunc);
extern unsigned long mvme166_keyb_init(unsigned long);
extern int mvme166_kbdrate (struct kbd_repeat *);
extern unsigned long mvme166_gettimeoffset (void);
extern void mvme166_gettod (int *year, int *mon, int *day, int *hour,
                           int *min, int *sec);
extern int mvme166_hwclk (int, struct hwclk_time *);
extern int mvme166_set_clock_mmss (unsigned long);
extern void mvme166_check_partition (struct gendisk *hd, unsigned int dev);
extern void mvme166_mksound( unsigned int count, unsigned int ticks );
extern void mvme166_reset (void);
extern void mvme166_waitbut(void);
extern struct consw mvme166_con;

int bcd2int (unsigned char b);

/* Save tick handler routine pointer, will point to do_timer() in
 * kernel/sched.c, called via mvme166_process_int() */

isrfunc (*tick_handler)();

int mvme166_kbdrate (struct kbd_repeat *k)
{
	printk ("RGH: Called mvme166_kbdrate\n");
	return 0;
}

void mvme166_mksound( unsigned int count, unsigned int ticks )
{
	printk ("RGH: Called mvme166_mksound\n");
}

void mvme166_reset()
{
	printk ("\r\n\nRGH: Called mvme166_reset\r\n"
			"\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r");
	/* The string of returns is to delay the reset until the whole
	 * message is output.  Assert reset bit in GCSR */
	*(volatile char *)0xfff40107 = 0x80;
}



void config_mvme166(void)
{
    mach_process_int	 = mvme166_process_int;
    mach_sched_init      = mvme166_sched_init;
    mach_keyb_init       = mvme166_keyb_init;
    mach_kbdrate         = mvme166_kbdrate;
    mach_init_INTS       = mvme166_init_INTS;
    mach_gettimeoffset   = mvme166_gettimeoffset;
    mach_gettod  	 = mvme166_gettod;
    mach_hwclk           = mvme166_hwclk;
    mach_set_clock_mmss	 = mvme166_set_clock_mmss;
    mach_check_partition = mvme166_check_partition;
    mach_mksound         = mvme166_mksound;
    mach_reset		 = mvme166_reset;
    conswitchp           = &mvme166_con;
}

extern long time_finetune;	/* from kernel/sched.c */

void mvme166_sched_init (isrfunc timer_routine)
{
    tick_handler = timer_routine;
    time_finetune = 0;
    /* Using vme2chip tick timer 1 */
    *(volatile unsigned long *)0xfff40054 = 0;
    *(volatile unsigned long *)0xfff40050 = 10000;	/* 10ms */
    *(volatile unsigned long *)0xfff40060 |= 3;
    *(volatile unsigned long *)0xfff4006c |= 0x01000000;
    *(volatile unsigned long *)0xfff40078 |= 0x00000006;
}

/* This is always executed with interrupts disabled.  */
unsigned long mvme166_gettimeoffset (void)
{
    return (*(volatile unsigned long *)0xfff40054);
}

extern void mvme166_gettod (int *year, int *mon, int *day, int *hour,
                           int *min, int *sec)
{
	rtc->ctrl = RTC_READ;
	*year = bcd2int (rtc->bcd_year);
	*mon = bcd2int (rtc->bcd_mth);
	*day = bcd2int (rtc->bcd_dom);
	*hour = bcd2int (rtc->bcd_hr);
	*min = bcd2int (rtc->bcd_min);
	*sec = bcd2int (rtc->bcd_sec);
	rtc->ctrl = 0;
}

int bcd2int (unsigned char b)
{
	return ((b>>4)*10 + (b&15));
}

int mvme166_hwclk(int op, struct hwclk_time *t)
{
	return 0;
}

int mvme166_set_clock_mmss (unsigned long nowtime)
{
	return 0;
}
