/*
 * atari_scsi.h -- Header file for the Atari native SCSI driver
 *
 * Copyright 1994 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 *
 * (Loosely based on the work of Robert De Vries' team)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */


#ifndef ATARI_SCSI_H
#define ATARI_SCSI_H

/* (I_HAVE_OVERRUNS stuff removed) */

#ifndef ASM
int atari_scsi_abort (Scsi_Cmnd *);
int atari_scsi_detect (Scsi_Host_Template *);
const char *atari_scsi_info (struct Scsi_Host *);
int atari_scsi_queue_command (Scsi_Cmnd *, void (*done) (Scsi_Cmnd *));
int atari_scsi_reset (Scsi_Cmnd *);

/* The values for CMD_PER_LUN and CAN_QUEUE are somehow arbitrary. Higher
 * values may work, too; try it! (but it costs memory!) */

/* But there seems to be a bug somewhere that requires CAN_QUEUE to be
 * 2*CMD_OER_LUN. At least on a TT, no suprious timeouts seen since
 * changed CMD_PER_LUN... */

/* Note: The Falcon currently uses hardcoded 2/1 setting due to other
 * problems... */

#ifndef ATARI_CMD_PER_LUN
#define ATARI_CMD_PER_LUN 8
#endif

#ifndef ATARI_CAN_QUEUE
#define ATARI_CAN_QUEUE 16
#endif

#ifdef HOSTS_C

#define ATARI_SCSI { NULL, NULL,					\
  "Atari native SCSI",							\
  atari_scsi_detect,							\
  /* release */ NULL,							\
  atari_scsi_info,								\
  /* command */ NULL,							\
  atari_scsi_queue_command,						\
  atari_scsi_abort,								\
  atari_scsi_reset,								\
  /* slave_attach */ NULL,						\
  /* bios_param */ NULL,						\
  /* can queue */ 			ATARI_CAN_QUEUE,		\
  /* host_id */				7,					\
  /* scatter gather */ 		SG_ALL,				\
  /* cmd per lun */ 		ATARI_CMD_PER_LUN,		\
  /* present */ 			0,					\
  /* unchecked ISA DMA */	0,					\
  /* use_clustering */		DISABLE_CLUSTERING }

#else

#define NCR5380_implementation_fields 	   /* none */
#define NCR5380_local_declare() int __none /* bad design */
#define NCR5380_setup(instance) __none = 0 /* to avoid warning */

#define NCR5380_read(reg)		  atari_scsi_reg_read( reg )
#define NCR5380_write(reg, value) atari_scsi_reg_write( reg, value )

#define NCR5380_intr atari_scsi_intr
#define NCR5380_queue_command atari_scsi_queue_command
#define NCR5380_abort atari_scsi_abort
/* #define NCR5380_reset atari_scsi_reset */
#define NCR5380_dma_read_setup(inst,d,c) atari_scsi_dma_setup (inst, d, c, 0)
#define NCR5380_dma_write_setup(inst,d,c) atari_scsi_dma_setup (inst, d, c, 1)
#define NCR5380_dma_residual(inst) atari_scsi_dma_residual( inst )
/* #define NCR5380_pread(instance, d, c) atari_scsi_pread (instance, d, c) */
/* #define NCR5380_pwrite(instance, d, c) atari_scsi_pwrite (instance, d, c) */
#define	NCR5380_dma_xfer_len(i,cmd) atari_dma_xfer_len(cmd->SCp.this_residual)

#endif /* else def HOSTS_C */
#endif /* ndef ASM */
#endif /* ATARI_SCSI_H */


