/*
 *	mvme166_scsi.h  low level driver for ncr 53c710 on mvme166_scsi brd
 *	Copyright (C) 1995 R.G.Hirst, 01 Oct 95
 *
 *  based on:
 *
 *	seagate.h Copyright (C) 1992 Drew Eckhardt 
 *	low level scsi driver header for ST01/ST02 by
 *		Drew Eckhardt 
 *
 *	<drew@colorado.edu>
 */

#ifndef _MVME166_SCSI_H
	#define MVME166_SCSI_H
/*
	$Header
*/
#ifndef ASM
int mvme166_scsi_detect(Scsi_Host_Template *);
int mvme166_scsi_command(Scsi_Cmnd *);
int mvme166_scsi_queue_command(Scsi_Cmnd *, void (*done)(Scsi_Cmnd *));

int mvme166_scsi_abort(Scsi_Cmnd *);
const char *mvme166_scsi_info(struct Scsi_Host *);
int mvme166_scsi_reset(Scsi_Cmnd *); 

#ifndef NULL
	#define NULL 0
#endif

int mvme166_scsi_biosparam(Disk *, int, int*);

#define MVME166_SCSI  {NULL, NULL, "MVME166-LLD", mvme166_scsi_detect, 	\
			 NULL,				\
			 mvme166_scsi_info, mvme166_scsi_command,  	\
			 mvme166_scsi_queue_command, mvme166_scsi_abort, \
			 mvme166_scsi_reset, NULL, mvme166_scsi_biosparam, \
			 1, 7, SG_ALL, 1, 0, 0, DISABLE_CLUSTERING}
#endif

/*
	CONTROL defines
*/

#define CMD_ACK			0x40
#define CMD_BSY			0x20
#define CMD_SEL			0x10
#define CMD_ATN			0x08

/*
	STATUS
*/

#define STAT_REQ		0x80
#define STAT_BSY		0x20
#define STAT_MSG		0x04
#define STAT_CD			0x02
#define STAT_IO			0x01

/* 
	REQUESTS
*/

#define REQ_MASK (STAT_CD |  STAT_IO | STAT_MSG)
#define REQ_DATAOUT 0
#define REQ_DATAIN STAT_IO
#define REQ_CMDOUT STAT_CD
#define REQ_STATIN (STAT_CD | STAT_IO)
#define REQ_MSGOUT (STAT_MSG | STAT_CD)
#define REQ_MSGIN (STAT_MSG | STAT_CD | STAT_IO)

/*
	NCR CHIP REGISTER BITS
*/

#define	NCR_SCNTL1_ADB		0x40
#define NCR_SCNTL1_CON		0x10
#define NCR_SCNTL1_RST		0x08

#define NCR_ISTAT_RST		0x40

#define NCR_DCNTL_CF		0xc0	/* This value for 33MHz */
#define	NCR_DCNTL_EA		0x20
#define NCR_DCNTL_LLM		0x08
#define	NCR_DCNTL_COM		0x01

#define NCR_SCNTL0_EPG		0x04


extern volatile int mvme166_scsi_timeout;

/*
	Debugging code
*/

#define PHASE_BUS_FREE 1
#define PHASE_ARBITRATION 2
#define PHASE_SELECTION 4
#define PHASE_DATAIN 8 
#define PHASE_DATAOUT 0x10
#define PHASE_CMDOUT 0x20
#define PHASE_MSGIN 0x40
#define PHASE_MSGOUT 0x80
#define PHASE_STATUSIN 0x100
#define PHASE_ETC (PHASE_DATAIN | PHASE_DATA_OUT | PHASE_CMDOUT | PHASE_MSGIN | PHASE_MSGOUT | PHASE_STATUSIN)
#define PRINT_COMMAND 0x200
#define PHASE_EXIT 0x400
#define PHASE_RESELECT 0x800
#define DEBUG_FAST 0x1000
#define DEBUG_SG   0x2000
#define DEBUG_LINKED	0x4000
#define DEBUG_BORKEN	0x8000

/* 
 *	Control options - these are timeouts specified in .01 seconds.
 */

#define MVME166_SCSI_BUS_FREE_DELAY 25
#define MVME166_SCSI_SELECTION_DELAY 15

typedef unsigned char	BYTE;
typedef unsigned long	ULONG;

typedef struct {
	BYTE	sien,
		sdid,
		scntl1,
		scntl0,
		socl,
		sodl,
		sxfer,
		scid,
		sbcl,
		sbdl,
		sidl,
		sfbr,
		sstat2,
		sstat1,
		sstat0,
		dstat;
	ULONG	dsa;
	BYTE	ctest3,
		ctest2,
		ctest1,
		ctest0;
	ULONG	temp;
	BYTE	lcrc,
		ctest8,
		istat,
		dfifo;
	ULONG	dcmd_dbc,
		dnad,
		dsp,
		dsps,
		scratch;
	BYTE	dcntl,
		dwt,
		dien,
		dmode;
	ULONG	adder;
} NCR710;

#define NCR710_BASE	0xfff47000



#endif

