/*
** linux/amigatypes.h -- Types used in Amiga Linux kernel source
**
** Copyright 1992 by Greg Harp
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
** Created 09/29/92 by Greg Harp
*/

#ifndef _LINUX_AMIGATYPES_H_
#define _LINUX_AMIGATYPES_H_

#ifdef __KERNEL__ /* only if compiling the kernel */
#include <linux/types.h>
#endif

struct Node {
    struct  Node *ln_Succ;	/* Pointer to next (successor) */
    struct  Node *ln_Pred;	/* Pointer to previous (predecessor) */
    u_char   ln_Type;
    char    ln_Pri;		/* Priority, for sorting */
    char    *ln_Name;		/* ID string, null terminated */
};

/* defined board manufacturers */
/* ++Geert: Please update amiga/zorro.c if you make changes here */
/*          Many IDs were obtained by using ExpName V1.4 ((C) Richard Krber) */

#define MANUF_COMMODORE      (0x0202)	/* Commodore USA */
#define PROD_A2090A          (0x01)	/* CBM A2090/A2090A HD controller */
#define PROD_A2091           (0x02)	/* CBM A2091 HD controller */
#define PROD_A590            (0x03)	/* CBM A590 HD controller */
#define PROD_A2090B          (0x04)	/* CBM A2090B 2090 Autoboot card */
#define PROD_ARCNET          (0x09)	/* CBM Arcnet card */
#define PROD_CBMRAM          (0x0A)	/* CBM A2052/58.RAM | 590/2091.RAM */
#define PROD_A560RAM         (0x20)	/* CBM A560 memory module */
#define PROD_A2232PROTO      (0x45)	/* CBM A2232 serial prototype */
#define PROD_A2232           (0x46)	/* CBM A2232 serial production */
#define PROD_A2620           (0x50)	/* CBM A2620 68020/RAM card */
#define PROD_A2630           (0x51)	/* CBM A2630 68030/RAM card */
#define PROD_A4091           (0x54)	/* CBM A4091 SCSI HD controller */
#define PROD_ROMULATOR       (0x60)	/* CBM Romulator card */
#define PROD_A3000TESTFIX    (0x61)	/* CBM A3000 test fixture */
#define PROD_A2065           (0x70)	/* CBM A2065 Ethernet card */

#define MANUF_COMMODORE2     (0x0201)	/* Commodore Germany */
#define PROD_BRIDGEBOARD     (0x01)	/* CBM Bridgeboard */
#define PROD_A2386SX         (0x67)	/* CBM A2386-SX Bridgeboard */

#define MANUF_U_LOWELL       (0x0406)	/* University of Lowell */
#define PROD_A2410           (0x00)	/* CBM A2410 Hires Graphics card */

#define MANUF_AMERISTAR      (0x041D)	/* Ameristar */
#define PROD_AMERISTAR2065   (0x01)	/* Ethernet card */

#define MANUF_GVP            (0x07E1)	/* Great Valley Products */
#define PROD_GVPIISCSI       (0x0B)	/* GVP Series II SCSI */
#define PROD_GVPIISCSI_2     (0x09)	/* evidence that the driver works
					   for this product code also */
#define PROD_GVPIIRAM        (0x0A)	/* GVP Series II RAM */
#define PROD_GVPIV_24        (0x20)	/* GVP IV-24 Graphics Board */
#define PROD_GVPIO_EXT       (0xFF)	/* GVP I/O Extender */

#define MANUF_GVP2           (0x0891)	/* Great Valley Products */
#define PROD_SPECTRUM_RAM    (0x01)	/* GVP Spectrum Graphics Board */
#define PROD_SPECTRUM_REG    (0x02)

#define MANUF_VILLAGE_TRONIC (0x0877)	/* Village Tronic */
#define PROD_ARIADNE         (0xC9)	/* Ariadne Ethernet */
#define PROD_PICASSO_II_RAM  (0x0B)	/* Picasso II Graphics Board */
#define PROD_PICASSO_II_REG  (0x0C)

#define MANUF_HELFRICH1      (0x0861)	/* Helfrich */
#define PROD_RAINBOW3        (0x21)	/* Rainbow3 Graphics Board */

#define MANUF_HELFRICH2      (0x0893)
#define PROD_PICCOLO_RAM     (0x05)	/* Piccolo Graphics Board */
#define PROD_PICCOLO_REG     (0x06)
#define PROD_PEGGY_PLUS      (0x07)	/* PeggyPlus MPEG Decoder Board */
#define PROD_SD64_RAM        (0x0A)	/* SD64 Graphics Board */
#define PROD_SD64_REG        (0x0B)

#define MANUF_HYDRA_SYSTEMS  (0x0849)	/* Hydra Systems */
#define PROD_AMIGANET        (0x01)	/* Amiganet Board */

#define MANUF_MEMPHIS        (0x0100)	/* Memphis */
#define PROD_STORMBRINGER    (0x00)	/* Stormbringer */

#define MANUF_SUPRA          (0x0420)	/* Supra */
#define PROD_SUPRA_2400MODEM (0x10)	/* Supra 2400 Modem */

#define MANUF_BSC            (0x07FE)
#define PROD_ALF_3_SCSI      (0x03)	/* BSC ALF 3 SCSI Controller */

#define MANUF_BSC2           (0x082C)
#define PROD_OKTAGON_2008    (0x05)	/* BSC Oktagon 2008 SCSI Controller */
#define PROD_TANDEM          (0x06)	/* BSC Tandem */
#define PROD_PAR_SER_CARD    (0x10)	/* BSC Par/Ser Card */
#define PROD_MULTIFACE_III   (0x12)	/* BSC MultiFace III */
#define PROD_ISDN_MASTER     (0x40)	/* BSC ISDN Master */

#define MANUF_JOCHHEIM       (0x0804)	/* Jochheim */
#define PROD_JOCHHEIM_RAM    (0x01)	/* Jochheim RAM */

#define MANUF_CHECKPOINT     (0x0807)	/* Checkpoint Technologies */
#define PROD_SERIAL_SOLUTION (0x00)	/* Serial Solution */

#define MANUF_GOLEM          (0x0819)	/* Golem */
#define PROD_GOLEM_SCSI_II   (0x02)	/* Golem SCSI-II Controller */

#define MANUF_ADV_SYS_SOFT   (0x0836)	/* Advanced Systems & Software */
#define PROD_NEXUS_SCSI      (0x01)	/* Nexus SCSI Controller */
#define PROD_NEXUS_RAM       (0x08)	/* Nexus RAM */

#define MANUF_XPERT_PRODEV   (0x0845)	/* XPert/ProDev */
#define PROD_MERLIN_RAM      (0x03)	/* Merlin Graphics Board */
#define PROD_MERLIN_REG      (0x04)

#define MANUF_UTILITIES_ULTD (0x087B)	/* Utilities Unlimited */
#define PROD_EMPLANT_DELUXE  (0x15)	/* Emplant Deluxe */
#define PROD_EMPLANT_DELUXE2 (0x20)	/* Emplant Deluxe */

#define MANUF_HARMS_PROF     (0x0A00)	/* Harms Professional */
#define PROD_3500_TURBO      (0xd0)	/* 3500 Turbo board */

#define MANUF_VORTEX         (0x2017)	/* Vortex */
#define PROD_GOLDEN_GATE_386 (0x07)	/* Golden Gate 80386 Board */
#define PROD_GOLDEN_GATE_RAM (0x08)	/* Golden Gate RAM */
#define PROD_GOLDEN_GATE_486 (0x09)	/* Golden Gate 80486 Board */

#define MANUF_PHASE5         (0x2140)	/* Phase5 */
#define PROD_FASTLANE_RAM    (0x0a)	/* FastLane RAM */
#define PROD_FASTLANE_SCSI   (0x0b)	/* FastLane SCSI */
#define PROD_CYBERVISION     (0x22)	/* CyberVision64 Graphics Board */

#define MANUF_APOLLO         (0x2222)	/* Apollo */
#define PROD_AT_APOLLO       (0x22)	/* AT-Apollo */

#define MANUF_UWE_GERLACH    (0x3FF7)	/* Uwe Gerlach */
#define PROD_UG_RAM_ROM      (0xd4)	/* RAM/ROM */

#define MANUF_MACROSYSTEMS   (0x089B)	/* MacroSystems USA */
#define PROD_WARP_ENGINE     (0x13)	/* Warp Engine */

#define MANUF_MACROSYSTEMS2  (0x4754)	/* MacroSystems Germany */
#define PROD_MAESTRO         (0x03)	/* Maestro */
#define PROD_VLAB            (0x04)	/* VLab */
#define PROD_MAESTRO_PRO     (0x05)	/* Maestro Pro */
#define PROD_RETINA_Z2       (0x06)	/* Retina Z2 Graphics Board */
#define PROD_MULTI_EVOLUTION (0x08)	/* MultiEvolution */

#define MANUF_MICROBOTICS    (0x03F2)	/* MicroBotics */
#define PROD_VXL_30          (0x45)	/* VXL-30 Turbo Board */

#define MANUF_HACKER         (0x07db)	/* Test only: no product definitions */


struct ExpansionRom {
    /* -First 16 bytes of the expansion ROM */
    u_char	er_Type;	/* Board type, size and flags */
    u_char	er_Product;	/* Product number, assigned by manufacturer */
    u_char	er_Flags;	/* Flags */
    u_char	er_Reserved03;	/* Must be zero ($ff inverted) */
    u_short	er_Manufacturer; /* Unique ID,ASSIGNED BY COMMODORE-AMIGA! */
    u_long	er_SerialNumber; /* Available for use by manufacturer */
    u_short	er_InitDiagVec; /* Offset to optional "DiagArea" structure */
    u_char	er_Reserved0c;
    u_char	er_Reserved0d;
    u_char	er_Reserved0e;
    u_char	er_Reserved0f;
};

/* er_Type board type bits */
#define ERT_TYPEMASK	0xc0
#define ERT_ZORROII	0xc0
#define ERT_ZORROIII	0x80

struct ConfigDev {
    struct Node 	cd_Node;
    u_char		cd_Flags;	/* (read/write) */
    u_char		cd_Pad; 	/* reserved */
    struct ExpansionRom cd_Rom; 	/* copy of board's expansion ROM */
    void		*cd_BoardAddr;	/* where in memory the board was placed */
    u_long		cd_BoardSize;	/* size of board in bytes */
    u_short		cd_SlotAddr;	/* which slot number (PRIVATE) */
    u_short		cd_SlotSize;	/* number of slots (PRIVATE) */
    void		*cd_Driver;	/* pointer to node of driver */
    struct ConfigDev	*cd_NextCD;	/* linked list of drivers to config */
    u_long		cd_Unused[4];	/* for whatever the driver wants */
};

/*
 * Different models of Amiga
 */
#define AMI_UNKNOWN  (0)
#define AMI_500      (1)
#define AMI_2000     (2)
#define AMI_3000     (3)
#define AMI_4000     (4)
#define AMI_1200     (5)

/*
 * chipsets
 */
#define CS_STONEAGE (0)
#define CS_OCS	    (1)
#define CS_ECS	    (2)
#define CS_AGA	    (3)

/*
 * number of CIA Eclock ticks for a single scheduling tick (jiffy)
 */

#define NTSC_JIFFY_TICKS   (7159)
#define PAL_JIFFY_TICKS    (7094)

#endif /* asm-m68k/amigatypes.h */
