/*
** asm/bootinfo.h -- Definition of the Linux/68K boot information structure
**
** Copyright 1992 by Greg Harp
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
** Created 09/29/92 by Greg Harp
**
** 5/2/94 Roman Hodek:
**   Added bi_atari part of the machine dependant union bi_un; for now it
**	 contains just a model field to distinguish between TT and Falcon.
*/

#ifndef BOOTINFO_H
#define BOOTINFO_H

#include <asm/amigatypes.h>

/*
 * Amiga specific part of bootinfo structure.
 */

#define NUM_AUTO    16

#define NTSC_ECLOCK 715909
#define PAL_ECLOCK  709379

struct bi_Amiga {
    int
        model;				/* Amiga Model (3000?) */

/* ++Geert: The driver for AutoConfig device number i should set bit i of */
/*          autocon_configured so amiga/zorro.c knows about that. */
/*          Note that from the bootstrap viewpoint the addition of */
/*          autocon_configured is backwards compatible with the old bi_Amiga */
/*          structure (2*sizeof(u_short) == sizeof(int)). */

    unsigned short
	autocon_configured;		/* Bitmap of configured devices */

    unsigned short
	num_autocon;			/* # of autoconfig devices found */

    struct ConfigDev
	autocon[NUM_AUTO];		/* up to 16 autoconfig devices */

    unsigned long
	chip_size;			/* size of chip memory (bytes) */

    unsigned char
	vblank; 			/* VBLANK frequency */

    unsigned char
	psfreq; 			/* power supply frequency */

    unsigned long
	eclock; 			/* EClock frequency */

    unsigned long
	chipset;			/* native chipset present */
};


/* Atari specific part of bootinfo */

/*
 * Define several Hardware-Chips for indication so that for the ATARI we do
 * no longer decide whether it is a Falcon or other machine . It's just
 * important what hardware the machine uses
 */

/* ++roman 08/08/95: rewritten from ORing constants to a C bitfield */

#define ATARIHW_DECLARE(name)	unsigned name : 1
#define	ATARIHW_SET(name)	(boot_info.bi_atari.hw_present.name = 1)
#define	ATARIHW_PRESENT(name)	(boot_info.bi_atari.hw_present.name)

struct bi_Atari {
	struct {
		/* video hardware */
		ATARIHW_DECLARE(STND_SHIFTER);	/* ST-Shifter - no base low ! */   
		ATARIHW_DECLARE(EXTD_SHIFTER);	/* STe-Shifter - 24 bit adress */
		ATARIHW_DECLARE(TT_SHIFTER);	/* TT-Shifter */
		ATARIHW_DECLARE(VIDEL_SHIFTER);	/* Falcon-Shifter */
		/* sound hardware */
		ATARIHW_DECLARE(YM_2149);	/* Yamaha YM 2149 */
		ATARIHW_DECLARE(PCM_8BIT);	/* PCM-Sound in STe-ATARI */
		ATARIHW_DECLARE(CODEC);		/* CODEC Sound (Falcon) */
		/* disk storage interfaces */
		ATARIHW_DECLARE(TT_SCSI);	/* Directly mapped NCR5380 */
		ATARIHW_DECLARE(ST_SCSI);	/* NCR5380 via ST-DMA (Falcon) */
		ATARIHW_DECLARE(ACSI);		/* Standard ACSI like in STs */
		ATARIHW_DECLARE(IDE);		/* IDE Interface */
		ATARIHW_DECLARE(FDCSPEED);	/* 8/16 MHz switch for FDC */
		/* other I/O hardware */
		ATARIHW_DECLARE(ST_MFP);	/* The ST-MFP (there should
						   be no Atari without
						   it... but who knows?) */
		ATARIHW_DECLARE(TT_MFP);	/* 2nd MFP */		      
		ATARIHW_DECLARE(SCC);		/* Serial Communications Contr. */
		ATARIHW_DECLARE(ST_ESCC);	/* SCC Z83230 in an ST */
		ATARIHW_DECLARE(ANALOG_JOY);	/* Paddle Interface for STe
						   and Falcon */
		ATARIHW_DECLARE(MICROWIRE);	/* Microwire Interface */
		/* DMA */
		ATARIHW_DECLARE(STND_DMA);	/* 24 Bit limited ST-DMA */
		ATARIHW_DECLARE(EXTD_DMA);	/* 32 Bit ST-DMA */
		ATARIHW_DECLARE(SCSI_DMA);	/* DMA for the NCR5380 */
		ATARIHW_DECLARE(SCC_DMA);	/* DMA for the SCC */
		/* real time clocks */
		ATARIHW_DECLARE(TT_CLK);	/* TT compatible clock chip */
		ATARIHW_DECLARE(MSTE_CLK);	/* Mega ST(E) clock chip */
		/* supporting hardware */
		ATARIHW_DECLARE(SCU);		/* System Control Unit */
		ATARIHW_DECLARE(BLITTER);	/* Blitter */
		ATARIHW_DECLARE(VME);		/* VME Bus */
	} hw_present;
	unsigned long mch_cookie;	/* _MCH cookie from TOS */
};

/*
 * CPU and FPU types
 */
#define CPU_68020    (1)
#define CPU_68030    (2)
#define CPU_68040    (4)
#define CPU_68060    (8)
#define CPU_MASK     (31)
#define FPU_68881    (32)
#define FPU_68882    (64)
#define FPU_68040    (128)	/* Internal FPU */
#define FPU_68060    (256)	/* Internal FPU */

struct mem_info {
    unsigned long addr;		/* physical address of memory chunk */
    unsigned long size;		/* length of memory chunk (in bytes) */
};

#define NUM_MEMINFO  4

#define MACH_AMIGA   1
#define MACH_ATARI   2
#define MACH_MAC     3
#define MACH_MVME166 4

#define CL_SIZE      (256)

struct bootinfo {
    unsigned long
	machtype;			/* machine type */

    unsigned long
	cputype;			/* system CPU & FPU */

    struct mem_info
	memory[NUM_MEMINFO];		/* memory description */

    int
	num_memory;			/* # of memory blocks found */

    unsigned long
	ramdisk_size;			/* ramdisk size in 1024 byte blocks */

    unsigned long
	ramdisk_addr;			/* address of the ram disk in mem */

    char
        command_line[CL_SIZE];		/* kernel command line parameters */

    union {
	struct bi_Amiga bi_ami; 	/* Amiga specific information */
	struct bi_Atari bi_ata; 	/* Atari specific information */
    } bi_un;
};
#define bi_amiga bi_un.bi_ami
#define bi_atari bi_un.bi_ata
#define bi_mac	 bi_un.bi_mac

extern struct bootinfo
    boot_info;

#endif /* BOOTINFO_H */
