#ifndef _LINUX_LP_H
#define _LINUX_LP_H

/*
 * usr/include/linux/lp.h  modified for Amiga by Michael Rausch
 * modified for Atari by Andreas Schwab
 * bug fixed by Jes Sorensen 18/8-94:
 *     It was not possible to compile the kernel only for Atari or Amiga.
 *
 * linux i386 version  c.1991-1992 James Wiegand
 * many modifications copyright (C) 1992 Michael K. Johnson
 * Interrupt support added 1993 Nigel Gamble
 */

#include <linux/autoconf.h>
#include <asm/bootinfo.h>

/*
 * Per POSIX guidelines, this module reserves the LP and lp prefixes
 * These are the lp_table[minor].flags flags...
 */
#define LP_EXIST 0x0001
#define LP_BUSY	 0x0004
#define LP_ABORT 0x0040
#define LP_CAREFUL 0x0080
#define LP_ABORTOPEN 0x0100

/* timeout for each character.  This is relative to bus cycles -- it
 * is the count in a busy loop.  THIS IS THE VALUE TO CHANGE if you
 * have extremely slow printing, or if the machine seems to slow down
 * a lot when you print.  If you have slow printing, increase this
 * number and recompile, and if your system gets bogged down, decrease
 * this number.  This can be changed with the tunelp(8) command as well.
 */

#define LP_INIT_CHAR 1000

/* The parallel port specs apparently say that there needs to be
 * a .5usec wait before and after the strobe.  Since there are wildly
 * different computers running linux, I can't come up with a perfect
 * value, but since it worked well on most printers before without,
 * I'll initialize it to 0.
 */

#define LP_INIT_WAIT 0

/* This is the amount of time that the driver waits for the printer to
 * catch up when the printer's buffer appears to be filled.  If you
 * want to tune this and have a fast printer (i.e. HPIIIP), decrease
 * this number, and if you have a slow printer, increase this number.
 * This is in hundredths of a second, the default 2 being .05 second.
 * Or use the tunelp(8) command, which is especially nice if you want
 * change back and forth between character and graphics printing, which
 * are wildly different...
 */

#define LP_INIT_TIME 2

/* IOCTL numbers */
#define LPCHAR   0x0601  /* corresponds to LP_INIT_CHAR */
#define LPTIME   0x0602  /* corresponds to LP_INIT_TIME */
#define LPABORT  0x0604  /* call with TRUE arg to abort on error,
			    FALSE to retry.  Default is retry.  */
#define LPSETIRQ 0x0605  /* call with new IRQ number,
			    or 0 for polling (no IRQ) */
#define LPGETIRQ 0x0606  /* get the current IRQ number */
#define LPWAIT   0x0608  /* corresponds to LP_INIT_WAIT */
#define LPCAREFUL   0x0609  /* call with TRUE arg to require out-of-paper, off-
			    line, and error indicators good on all writes,
			    FALSE to ignore them.  Default is ignore. */
#define LPABORTOPEN 0x060a  /* call with TRUE arg to abort open() on error,
			    FALSE to ignore error.  Default is ignore.  */
#define LPGETSTATUS 0x060b  /* return LP_S(minor) */
#define LPRESET     0x060c  /* reset printer */

/* timeout for printk'ing a timeout, in jiffies (100ths of a second).
   This is also used for re-checking error conditions if LP_ABORT is
   not set.  This is the default behavior. */

#define LP_TIMEOUT_INTERRUPT	(60 * HZ)
#define LP_TIMEOUT_POLLED	(10 * HZ)

#define LP_F	lp_table.flags		/* flags for busy, etc. */
#define LP_CHAR	lp_table.chars		/* busy timeout */
#define LP_TIME	lp_table.time		/* wait time */
#define LP_WAIT	lp_table.wait		/* strobe wait */
#define LP_IRQ	lp_table.irq		/* interrupt # */
					/* 0 means polled */

#define LP_BUFFER_SIZE 256

struct lp_struct {
	unsigned int irq;
	int flags;
	unsigned int chars;
	unsigned int time;
	unsigned int wait;
	struct wait_queue *lp_wait_q;
	char *lp_buffer;
};

struct lp_struct lp_table = {
	0, 0, LP_INIT_CHAR, LP_INIT_TIME, LP_INIT_WAIT, NULL, NULL,
}; 
#define LP_NO 1


#define AMIGA_LP_OUT(c)	ciaa.prb = c

/*  busy is low-active, and so are the cia bits */
#define AMIGA_LP_IS_BUSY()	(ciab.pra & 1)
#define AMIGA_LP_HAS_POUT()	(ciab.pra & 2)
#define AMIGA_LP_IS_ONLINE()	(ciab.pra & 4)
#define AMIGA_LP_HAS_ERROR()    ((ciab.pra & 3) || !(ciab.pra & 4))  /* FEF */

/* LP_OUT must be executed with interrupts blocked */
#define ATARI_LP_OUT(c)						\
{								\
   int wait = 0;						\
   sound_ym.rd_data_reg_sel = 15;				\
   sound_ym.wd_data = (c);					\
   sound_ym.rd_data_reg_sel = 14;				\
   while (wait != LP_WAIT) wait++;				\
   sound_ym.wd_data = sound_ym.rd_data_reg_sel & ~(1 << 5);	\
   while (wait) wait--;						\
   sound_ym.wd_data = sound_ym.rd_data_reg_sel | (1 << 5);	\
}

#define ATARI_LP_IS_BUSY()	(mfp.par_dt_reg & 1)
#define ATARI_LP_HAS_POUT()	0
#define ATARI_LP_IS_ONLINE()	!ATARI_LP_IS_BUSY() /* or 1 XXX */
#define ATARI_LP_HAS_ERROR()    ATARI_LP_IS_BUSY()  /* FEF */

#if defined (CONFIG_AMIGA) + defined (CONFIG_ATARI) == 1

#ifdef CONFIG_AMIGA
#define LP_OUT(c) AMIGA_LP_OUT(c)
#define LP_IS_BUSY() AMIGA_LP_IS_BUSY()
#define LP_HAS_POUT() AMIGA_LP_HAS_POUT()
#define LP_IS_ONLINE() AMIGA_LP_IS_ONLINE()
#define LP_HAS_ERROR() AMIGA_LP_HAS_ERROR()
#endif
#ifdef CONFIG_ATARI
#define LP_OUT(c) ATARI_LP_OUT(c)
#define LP_IS_BUSY() ATARI_LP_IS_BUSY()
#define LP_HAS_POUT() ATARI_LP_HAS_POUT()
#define LP_IS_ONLINE() ATARI_LP_IS_ONLINE()
#define LP_HAS_ERROR() ATARI_LP_HAS_ERROR()
#endif

#else /* multi configuration */

static __inline__ void
LP_OUT (int c)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      AMIGA_LP_OUT (c);
      break;
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      ATARI_LP_OUT (c);
      break;
#endif
    }
}

static __inline__ int
LP_IS_BUSY (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_IS_BUSY ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_IS_BUSY ();
#endif
    default:
      return 0;
    }
}

static __inline__ int
LP_HAS_POUT (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_HAS_POUT ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_HAS_POUT ();
#endif
    default:
      return 0;
    }
}

static __inline__ int
LP_IS_ONLINE (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_IS_ONLINE ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_IS_ONLINE ();
#endif
    default:
      return 0;
    }
}

static __inline__ int
LP_HAS_ERROR (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_HAS_ERROR ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_HAS_ERROR ();
#endif
    default:
      return 0;
    }
}

#endif

/*
 * function prototypes
 */

extern long lp_init(long);

#endif
